#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "adaptivetimestepper.hh"

void IterationRegister::registerCount(FixedPointIterationCounter count) {
  totalCount += count;
}

void IterationRegister::registerFinalCount(FixedPointIterationCounter count) {
  finalCount = count;
}

void IterationRegister::reset() {
  totalCount = FixedPointIterationCounter();
  finalCount = FixedPointIterationCounter();
}

template <class Factory, class Updaters, class ErrorNorm>
AdaptiveTimeStepper<Factory, Updaters, ErrorNorm>::AdaptiveTimeStepper(
    Factory &factory, Dune::ParameterTree const &parset,
    std::shared_ptr<Nonlinearity> globalFriction, Updaters &current,
    double relativeTime, double relativeTau,
    std::function<void(double, Vector &)> externalForces,
    ErrorNorm const &errorNorm,
    std::function<bool(Updaters &, Updaters &)> mustRefine)
    : relativeTime_(relativeTime),
      relativeTau_(relativeTau),
      finalTime_(parset.get<double>("problem.finalTime")),
      factory_(factory),
      parset_(parset),
      globalFriction_(globalFriction),
      current_(current),
      externalForces_(externalForces),
      mustRefine_(mustRefine),
      errorNorm_(errorNorm) {}

template <class Factory, class Updaters, class ErrorNorm>
bool AdaptiveTimeStepper<Factory, Updaters, ErrorNorm>::reachedEnd() {
  return relativeTime_ >= 1.0;
}

template <class Factory, class Updaters, class ErrorNorm>
bool AdaptiveTimeStepper<Factory, Updaters, ErrorNorm>::coarsen() {
  bool didCoarsen = false;
  while (relativeTime_ + relativeTau_ < 1.0) {
    R2_ = { R1_.updaters.clone() };
    step(R2_, relativeTime_ + relativeTau_, relativeTau_);

    UpdatersWithCount C{ current_.clone() };
    step(C, relativeTime_, 2.0 * relativeTau_);

    if (!mustRefine_(C.updaters, R2_.updaters)) {
      R2_ = {};
      R1_ = C;
      relativeTau_ *= 2.0;
      didCoarsen = true;
    } else {
      break;
    }
  }
  return didCoarsen;
}

template <class Factory, class Updaters, class ErrorNorm>
void AdaptiveTimeStepper<Factory, Updaters, ErrorNorm>::refine() {
  while (true) {
    UpdatersWithCount F1{ current_.clone() };
    step(F1, relativeTime_, relativeTau_ / 2.0);
    UpdatersWithCount F2{ F1.updaters.clone() };
    step(F2, relativeTime_ + relativeTau_ / 2.0, relativeTau_ / 2.0);

    if (!mustRefine_(R1_.updaters, F2.updaters)) {
      break;
    } else {
      R1_ = F1;
      R2_ = F2;
      relativeTau_ /= 2.0;
    }
  }
}

template <class Factory, class Updaters, class ErrorNorm>
IterationRegister AdaptiveTimeStepper<Factory, Updaters, ErrorNorm>::advance() {
  if (R2_.updaters == Updaters()) {
    R1_ = { current_.clone() };
    step(R1_, relativeTime_, relativeTau_); // counting again upon restart
  } else {
    R1_ = R2_;
  }

  iterationRegister_.reset();
  if (!coarsen())
    refine();

  current_ = R1_.updaters;
  iterationRegister_.registerFinalCount(R1_.count);
  relativeTime_ += relativeTau_;

  return iterationRegister_;
}

template <class Factory, class Updaters, class ErrorNorm>
void AdaptiveTimeStepper<Factory, Updaters, ErrorNorm>::step(
    UpdatersWithCount &updatersAndCount, double rTime, double rTau) {
  updatersAndCount.count =
      MyCoupledTimeStepper(finalTime_, factory_, parset_, globalFriction_,
                           updatersAndCount.updaters, errorNorm_,
                           externalForces_).step(rTime, rTau);
  iterationRegister_.registerCount(updatersAndCount.count);
}

#include "adaptivetimestepper_tmpl.cc"
